@@ -323,7 +323,13 @@ static int create_ghost(struct ghost_file *gf, GhostFileEntry *gfe, struct cr_im
 		goto err;
 	}
 
-	snprintf(path + ret, sizeof(path) - ret, "/%s", gf->remap.rpath);
+	/* Add a '/' only if we have no at the end */
+	if (path[root_len-1] != '/') {
+		path[root_len++] = '/';
+		path[root_len] = '\0';
+	}
+
+	snprintf(path + root_len, sizeof(path) - root_len, "%s", gf->remap.rpath);
 	ret = -1;
 again:
 	if (S_ISFIFO(gfe->mode)) {
@@ -358,7 +364,7 @@ again:
 		goto err;
 	}
 
-	strcpy(gf->remap.rpath, path + root_len + 1);
+	strcpy(gf->remap.rpath, path + root_len);
 	pr_debug("Remap rpath is %s\n", gf->remap.rpath);
 
 	ret = -1;
@@ -371,18 +377,18 @@ err:
 }
 
 static inline void ghost_path(char *path, int plen,
-		struct reg_file_info *rfi, RemapFilePathEntry *rfe)
+		struct reg_file_info *rfi, RemapFilePathEntry *rpe)
 {
-	snprintf(path, plen, "%s.cr.%x.ghost", rfi->path, rfe->remap_id);
+	snprintf(path, plen, "%s.cr.%x.ghost", rfi->path, rpe->remap_id);
 }
 
 static int collect_remap_ghost(struct reg_file_info *rfi,
-		RemapFilePathEntry *rfe)
+		RemapFilePathEntry *rpe)
 {
 	struct ghost_file *gf;
 
 	list_for_each_entry(gf, &ghost_files, list)
-		if (gf->id == rfe->remap_id)
+		if (gf->id == rpe->remap_id)
 			goto gf_found;
 
 	/*
@@ -391,7 +397,7 @@ static int collect_remap_ghost(struct reg_file_info *rfi,
 	 * issues with cross-device links.
 	 */
 
-	pr_info("Opening ghost file %#x for %s\n", rfe->remap_id, rfi->path);
+	pr_info("Opening ghost file %#x for %s\n", rpe->remap_id, rfi->path);
 
 	gf = shmalloc(sizeof(*gf));
 	if (!gf)
@@ -408,7 +414,7 @@ static int collect_remap_ghost(struct reg_file_info *rfi,
 	if (!gf->remap.rpath)
 		return -1;
 	gf->remap.rpath[0] = 0;
-	gf->id = rfe->remap_id;
+	gf->id = rpe->remap_id;
 	list_add_tail(&gf->list, &ghost_files);
 
 gf_found:
@@ -418,7 +424,7 @@ gf_found:
 }
 
 static int open_remap_ghost(struct reg_file_info *rfi,
-					RemapFilePathEntry *rfe)
+					RemapFilePathEntry *rpe)
 {
 	struct ghost_file *gf = container_of(rfi->remap, struct ghost_file, remap);
 	GhostFileEntry *gfe = NULL;
@@ -427,7 +433,7 @@ static int open_remap_ghost(struct reg_file_info *rfi,
 	if (rfi->remap->rpath[0])
 		return 0;
 
-	img = open_image(CR_FD_GHOST_FILE, O_RSTR, rfe->remap_id);
+	img = open_image(CR_FD_GHOST_FILE, O_RSTR, rpe->remap_id);
 	if (!img)
 		goto err;
 
@@ -446,7 +452,7 @@ static int open_remap_ghost(struct reg_file_info *rfi,
 	if (S_ISDIR(gfe->mode))
 		strncpy(gf->remap.rpath, rfi->path, PATH_MAX);
 	else
-		ghost_path(gf->remap.rpath, PATH_MAX, rfi, rfe);
+		ghost_path(gf->remap.rpath, PATH_MAX, rfi, rpe);
 
 	if (create_ghost(gf, gfe, img))
 		goto close_ifd;
@@ -469,15 +475,15 @@ err:
 }
 
 static int collect_remap_linked(struct reg_file_info *rfi,
-		RemapFilePathEntry *rfe)
+		RemapFilePathEntry *rpe)
 {
 	struct file_remap *rm;
 	struct file_desc *rdesc;
 	struct reg_file_info *rrfi;
 
-	rdesc = find_file_desc_raw(FD_TYPES__REG, rfe->remap_id);
+	rdesc = find_file_desc_raw(FD_TYPES__REG, rpe->remap_id);
 	if (!rdesc) {
-		pr_err("Can't find target file %x\n", rfe->remap_id);
+		pr_err("Can't find target file %x\n", rpe->remap_id);
 		return -1;
 	}
 
@@ -497,8 +503,7 @@ static int collect_remap_linked(struct reg_file_info *rfi,
 	return 0;
 }
 
-static int open_remap_linked(struct reg_file_info *rfi,
-		RemapFilePathEntry *rfe)
+static int open_remap_linked(struct reg_file_info *rfi)
 {
 	if (root_ns_mask & CLONE_NEWUSER) {
 		int rfd;
@@ -550,47 +555,47 @@ static int collect_remap_dead_process(struct reg_file_info *rfi,
 
 struct remap_info {
 	struct list_head list;
-	RemapFilePathEntry *rfe;
+	RemapFilePathEntry *rpe;
 	struct reg_file_info *rfi;
 };
 
 static int collect_one_remap(void *obj, ProtobufCMessage *msg, struct cr_img *i)
 {
 	struct remap_info *ri = obj;
-	RemapFilePathEntry *rfe;
+	RemapFilePathEntry *rpe;
 	struct file_desc *fdesc;
 
-	ri->rfe = rfe = pb_msg(msg, RemapFilePathEntry);
+	ri->rpe = rpe = pb_msg(msg, RemapFilePathEntry);
 
-	if (!rfe->has_remap_type) {
-		rfe->has_remap_type = true;
+	if (!rpe->has_remap_type) {
+		rpe->has_remap_type = true;
 		/* backward compatibility with images */
-		if (rfe->remap_id & REMAP_GHOST) {
-			rfe->remap_id &= ~REMAP_GHOST;
-			rfe->remap_type = REMAP_TYPE__GHOST;
+		if (rpe->remap_id & REMAP_GHOST) {
+			rpe->remap_id &= ~REMAP_GHOST;
+			rpe->remap_type = REMAP_TYPE__GHOST;
 		} else
-			rfe->remap_type = REMAP_TYPE__LINKED;
+			rpe->remap_type = REMAP_TYPE__LINKED;
 	}
 
-	fdesc = find_file_desc_raw(FD_TYPES__REG, rfe->orig_id);
+	fdesc = find_file_desc_raw(FD_TYPES__REG, rpe->orig_id);
 	if (fdesc == NULL) {
-		pr_err("Remap for non existing file %#x\n", rfe->orig_id);
+		pr_err("Remap for non existing file %#x\n", rpe->orig_id);
 		return -1;
 	}
 
 	ri->rfi = container_of(fdesc, struct reg_file_info, d);
 
-	switch (rfe->remap_type) {
+	switch (rpe->remap_type) {
 	case REMAP_TYPE__GHOST:
-		if (collect_remap_ghost(ri->rfi, ri->rfe))
+		if (collect_remap_ghost(ri->rfi, ri->rpe))
 			return -1;
 		break;
 	case REMAP_TYPE__LINKED:
-		if (collect_remap_linked(ri->rfi, ri->rfe))
+		if (collect_remap_linked(ri->rfi, ri->rpe))
 			return -1;
 		break;
 	case REMAP_TYPE__PROCFS:
-		if (collect_remap_dead_process(ri->rfi, rfe) < 0)
+		if (collect_remap_dead_process(ri->rfi, rpe) < 0)
 			return -1;
 		break;
 	default:
@@ -605,24 +610,24 @@ static int collect_one_remap(void *obj, ProtobufCMessage *msg, struct cr_img *i)
 static int prepare_one_remap(struct remap_info *ri)
 {
 	int ret = -1;
-	RemapFilePathEntry *rfe = ri->rfe;
+	RemapFilePathEntry *rpe = ri->rpe;
 	struct reg_file_info *rfi = ri->rfi;
 
-	pr_info("Configuring remap %#x -> %#x\n", rfi->rfe->id, rfe->remap_id);
+	pr_info("Configuring remap %#x -> %#x\n", rfi->rfe->id, rpe->remap_id);
 
-	switch (rfe->remap_type) {
+	switch (rpe->remap_type) {
 	case REMAP_TYPE__LINKED:
-		ret = open_remap_linked(rfi, rfe);
+		ret = open_remap_linked(rfi);
 		break;
 	case REMAP_TYPE__GHOST:
-		ret = open_remap_ghost(rfi, rfe);
+		ret = open_remap_ghost(rfi, rpe);
 		break;
 	case REMAP_TYPE__PROCFS:
 		/* handled earlier by collect_remap_dead_process */
 		ret = 0;
 		break;
 	default:
-		pr_err("unknown remap type %u\n", rfe->remap_type);
+		pr_err("unknown remap type %u\n", rpe->remap_type);
 		goto out;
 	}
 
@@ -692,11 +697,11 @@ int try_clean_remaps(bool only_ghosts)
 	int ret = 0;
 
 	list_for_each_entry(ri, &remaps, list) {
-		if (ri->rfe->remap_type == REMAP_TYPE__GHOST)
+		if (ri->rpe->remap_type == REMAP_TYPE__GHOST)
 			ret |= clean_one_remap(ri);
 		else if (only_ghosts)
 			continue;
-		else if (ri->rfe->remap_type == REMAP_TYPE__LINKED)
+		else if (ri->rpe->remap_type == REMAP_TYPE__LINKED)
 			ret |= clean_one_remap(ri);
 	}
 
@@ -965,18 +970,9 @@ int dead_pid_conflict(void)
 		if (!node)
 			continue;
 
-		if (node->state != TASK_THREAD) {
-			struct pstree_item *item;
-
-			/*
-			 * If the dead PID was given to a main thread of another
-			 * process, this is handled during restore.
-			 */
-			item = node->item;
-			if (item->pid->real == item->threads[i].real ||
-			    item->threads[i].ns[0].virt != pid)
-				continue;
-		}
+		/* Main thread */
+		if (node->state != TASK_THREAD)
+			continue;
 
 		pr_err("Conflict with a dead task with the same PID as of this thread (virt %d, real %d).\n",
 			node->ns[0].virt, node->real);
@@ -1365,7 +1361,7 @@ static int linkat_hard(int odir, char *opath, int ndir, char *npath, uid_t uid,
 
 	if (!( (errno == EPERM || errno == EOVERFLOW) && (root_ns_mask & CLONE_NEWUSER) )) {
 		errno_save = errno;
-		pr_warn("Can't link %s -> %s", opath, npath);
+		pr_warn("Can't link %s -> %s\n", opath, npath);
 		errno = errno_save;
 		return ret;
 	}
@@ -1587,6 +1583,7 @@ int open_path(struct file_desc *d,
 	struct reg_file_info *rfi;
 	char *orig_path = NULL;
 	char path[PATH_MAX];
+	int inh_fd = -1;
 
 	if (inherited_fd(d, &tmp))
 		return tmp;
@@ -1596,6 +1593,7 @@ int open_path(struct file_desc *d,
 	if (rfi->rfe->ext) {
 		tmp = inherit_fd_lookup_id(rfi->rfe->name);
 		if (tmp >= 0) {
+			inh_fd = tmp;
 			mntns_root = open_pid_proc(PROC_SELF);
 			snprintf(path, sizeof(path), "fd/%d", tmp);
 			orig_path = rfi->path;
@@ -1654,8 +1652,10 @@ ext:
 	tmp = open_cb(mntns_root, rfi, arg);
 	if (tmp < 0) {
 		pr_perror("Can't open file %s", rfi->path);
+		close_safe(&inh_fd);
 		return -1;
 	}
+	close_safe(&inh_fd);
 
 	if ((rfi->rfe->has_size || rfi->rfe->has_mode) &&
 	    !rfi->size_mode_checked) {
