@@ -45,7 +45,7 @@
  *  	contents modification especially when tasks are
  *  	migrating between different kernel versions
  *  - heap
- *  	"heap" area in application, currently for inforamtion only
+ *  	"heap" area in application, currently for information only
  *  - file private
  *  	stands for privately memory mapped files
  *  - file shared
@@ -103,6 +103,7 @@ extern bool img_common_magic;
 #define O_DUMP		(O_WRONLY | O_CREAT | O_TRUNC)
 #define O_SHOW		(O_RDONLY | O_NOBUF)
 #define O_RSTR		(O_RDONLY)
+#define O_FORCE_LOCAL	(O_SYNC)
 
 struct cr_img {
 	union {
