@@ -4,6 +4,9 @@
 #include <stdbool.h>
 #include "int.h"
 #include "vma.pb-c.h"
+#include "pid.h"
+#include "proc_parse.h"
+#include "inventory.pb-c.h"
 
 struct parasite_ctl;
 struct vm_area_list;
@@ -12,10 +15,14 @@ struct pstree_item;
 struct vma_area;
 
 struct mem_dump_ctl {
-	bool	pre_dump;
+	bool			pre_dump;
+	bool			lazy;
+	struct proc_pid_stat	*stat;
+	InventoryEntry		*parent_ie;
 };
 
 extern bool vma_has_guard_gap_hidden(struct vma_area *vma);
+extern bool page_is_zero(u64 pme);
 extern bool page_in_parent(bool dirty);
 extern int prepare_mm_pid(struct pstree_item *i);
 extern void prepare_cow_vmas(void);
