@@ -47,6 +47,11 @@ extern int move_veth_to_bridge(void);
 
 extern int kerndat_link_nsid(void);
 extern int net_get_nsid(int rtsk, int fd, int *nsid);
+extern struct ns_id *net_get_root_ns();
 extern int kerndat_nsid(void);
+extern void check_has_netns_ioc(int fd, bool *kdat_val, const char *name);
+extern int net_set_ext(struct ns_id *ns);
+extern struct ns_id *get_root_netns();
+extern int read_net_ns_img();
 
 #endif /* __CR_NET_H__ */
