@@ -54,4 +54,9 @@ static inline bool pid_rst_prio(unsigned pid_a, unsigned pid_b)
 	return pid_a < pid_b;
 }
 
+static inline bool pid_rst_prio_eq(unsigned pid_a, unsigned pid_b)
+{
+	return pid_a <= pid_b;
+}
+
 #endif /* __CR_PID_H__ */
