@@ -162,7 +162,8 @@ int parasite_dump_thread_leader_seized(struct parasite_ctl *ctl, int pid, CoreEn
 	return dump_thread_core(pid, core, args);
 }
 
-int parasite_dump_thread_seized(struct parasite_ctl *ctl, int id,
+int parasite_dump_thread_seized(struct parasite_thread_ctl *tctl,
+				struct parasite_ctl *ctl, int id,
 				struct pid *tid, CoreEntry *core)
 {
 	struct parasite_dump_thread *args;
@@ -171,7 +172,6 @@ int parasite_dump_thread_seized(struct parasite_ctl *ctl, int id,
 	CredsEntry *creds = tc->creds;
 	struct parasite_dump_creds *pc;
 	int ret;
-	struct parasite_thread_ctl *tctl;
 
 	BUG_ON(id == 0); /* Leader is dumped in dump_task_core_all */
 
@@ -180,10 +180,6 @@ int parasite_dump_thread_seized(struct parasite_ctl *ctl, int id,
 	pc = args->creds;
 	pc->cap_last_cap = kdat.last_cap;
 
-	tctl = compel_prepare_thread(ctl, pid);
-	if (!tctl)
-		return -1;
-
 	tc->has_blk_sigset = true;
 	memcpy(&tc->blk_sigset, compel_thread_sigmask(tctl), sizeof(k_rtsigset_t));
 	ret = compel_get_thread_regs(tctl, save_task_regs, core);
@@ -468,12 +464,55 @@ static int make_sigframe(void *arg, struct rt_sigframe *sf, struct rt_sigframe *
 	return construct_sigframe(sf, rtsf, bs, (CoreEntry *)arg);
 }
 
+static int parasite_prepare_threads(struct parasite_ctl *ctl,
+				    struct pstree_item *item)
+{
+	struct parasite_thread_ctl **thread_ctls;
+	uint64_t *thread_sp;
+	int i;
+
+	thread_ctls = xzalloc(sizeof(*thread_ctls) * item->nr_threads);
+	if (!thread_ctls)
+		return -1;
+
+	thread_sp = xzalloc(sizeof(*thread_sp) * item->nr_threads);
+	if (!thread_sp)
+		goto free_ctls;
+
+	for (i = 0; i < item->nr_threads; i++) {
+		struct pid *tid = &item->threads[i];
+
+		if (item->pid->real == tid->real) {
+			thread_sp[i] = compel_get_leader_sp(ctl);
+			continue;
+		}
+
+		thread_ctls[i] = compel_prepare_thread(ctl, tid->real);
+		if (!thread_ctls[i])
+			goto free_sp;
+
+		thread_sp[i] = compel_get_thread_sp(thread_ctls[i]);
+	}
+
+	dmpi(item)->thread_ctls = thread_ctls;
+	dmpi(item)->thread_sp = thread_sp;
+
+	return 0;
+
+free_sp:
+	xfree(thread_sp);
+free_ctls:
+	xfree(thread_ctls);
+	return -1;
+}
+
 struct parasite_ctl *parasite_infect_seized(pid_t pid, struct pstree_item *item,
 		struct vm_area_list *vma_area_list)
 {
 	struct parasite_ctl *ctl;
 	struct infect_ctx *ictx;
 	unsigned long p;
+	int ret;
 
 	BUG_ON(item->threads[0].real != pid);
 
@@ -487,6 +526,10 @@ struct parasite_ctl *parasite_infect_seized(pid_t pid, struct pstree_item *item,
 	if (!ctl)
 		return NULL;
 
+	ret = parasite_prepare_threads(ctl, item);
+	if (ret)
+		return NULL;
+
 	ictx = compel_infect_ctx(ctl);
 
 	ictx->open_proc = do_open_proc;
@@ -511,6 +554,8 @@ struct parasite_ctl *parasite_infect_seized(pid_t pid, struct pstree_item *item,
 		ictx->flags |= INFECT_NO_BREAKPOINTS;
 	if (kdat.compat_cr)
 		ictx->flags |= INFECT_COMPATIBLE;
+	if (kdat.x86_has_ptrace_fpu_xsave_bug)
+		ictx->flags |= INFECT_X86_PTRACE_MXCSR_BUG;
 
 	ictx->log_fd = log_get_fd();
 
@@ -530,4 +575,3 @@ struct parasite_ctl *parasite_infect_seized(pid_t pid, struct pstree_item *item,
 
 	return ctl;
 }
-
