@@ -24,6 +24,9 @@
 #include "protobuf.h"
 #include "images/sk-packet.pb-c.h"
 
+#undef  LOG_PREFIX
+#define LOG_PREFIX "skqueue: "
+
 struct sk_packet {
 	struct list_head	list;
 	SkPacketEntry		*entry;
@@ -126,7 +129,7 @@ static int dump_packet_cmsg(struct msghdr *mh, SkPacketEntry *pe)
 			if (n_rights) {
 				/*
 				 * Even if user is sending more than one cmsg with
-				 * rights, kernel merges them alltogether on recv.
+				 * rights, kernel merges them altogether on recv.
 				 */
 				pr_err("Unexpected 2nd SCM_RIGHTS from the kernel\n");
 				return -1;
