@@ -1,4 +1,3 @@
-#define _GNU_SOURCE
 #include <string.h>
 #include <fcntl.h>
 #include <unistd.h>
@@ -14,6 +13,12 @@
 #define CLONE_NEWNS     0x00020000
 #endif
 
+#ifdef REMAP_PID_ROOT
+const char *proc_path = "/proc/%d";
+#else
+const char *proc_path = "/proc/%d/mountinfo";
+#endif
+
 const char *test_doc	= "Check that dead pid's /proc entries are remapped correctly";
 const char *test_author	= "Tycho Andersen <tycho.andersen@canonical.com>";
 
@@ -30,17 +35,18 @@ int main(int argc, char **argv)
 	}
 
 	if (pid == 0) {
-		test_msg("child is %d\n", pid);
 		/* Child process just sleeps until it is killed. All we need
 		 * here is a process to open the mountinfo of. */
 		while(1)
 			sleep(10);
 	} else {
+		test_msg("child is %d\n", pid);
+
 		int fd, ret;
 		char path[PATH_MAX];
 		pid_t result;
 
-		sprintf(path, "/proc/%d", pid);
+		sprintf(path, proc_path, pid);
 		fd = open(path, O_RDONLY);
 		if (fd < 0) {
 			fail("failed to open fd");
